package tstego;


import java.awt.image.BufferedImage;
import java.util.LinkedList;
import java.util.List;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;

import conversion.BaseXToBase10;
import conversion.ModifyImage;
import conversion.RGBDifferenceToInteger;

/**
 * @author Shane
 *
 */
public class ThumbStego {

	/**
	 * The finals used for specifying which color channel to look at.
	 */
	private static final int R = 0, G=1, B=2;
	
	/**
	 * Returns whether the data can fit into the image or not.
	 * @param sourceImage
	 * @param dataSize the size of the data
	 * @param colorChannelOffset the plus minus offset for the pixels (1 = +-1) (Fixed at 3 usually)
	 * @return true is returned if the picture can hold the message
	 */
	public static boolean checkSize(BufferedImage sourceImage, int dataSize, int colorChannelOffset){
		if(dataSize < findNumModifiablePixels(sourceImage,colorChannelOffset))
			return true;
		return false;
	}
	
	/**
	 * Returns whether the data can fit into the image or not.
	 * @param sourceImage
	 * @param data the data to check to see
	 * @param colorChannelOffset the plus minus offset for the pixels (1 = +-1) (Fixed at 3 usually)
	 * @return true is returned if the picture can hold the message
	 */
	public static boolean checkSize(BufferedImage sourceImage, List data, int colorChannelOffset){
		return checkSize(sourceImage, data.size(), colorChannelOffset);
	}
	
	
	/**
	 * The general one which sets the colorChannelOffset to 3, no point using anything but
	 * 3, just left for easy recoding later (changing the base).
	 * @param sourceImage The image to have the data encrypted in
	 * @param data The file the data is in to be encryped
	 * @param colorChannelOffset
	 * @return The encrypted image is returned.
	 */
	public static BufferedImage embedData(BufferedImage sourceImage, File data, int colorChannelOffset){

		/**
		 * This reads in the file to be encrypted in the image
		 */
		File fileToBeEncrypted;
		FileInputStream fis = null;
		try{
			fileToBeEncrypted = data;
			fis = new FileInputStream(fileToBeEncrypted);
		}catch(Exception e){
			System.err.println("There was a problem with the file being encrypted.");
		}
		

		LinkedList<Integer> dataList = new LinkedList<Integer>();
		
		/**
		 * This adds all the data from the file into a LinkedList
		 * byte by byte.
		 */
		try{
			int temp;
			while((temp = fis.read())!=-1)
				dataList.add(temp);
		}catch(IOException e){
			System.err.println("There was an error with your file to be encrypted in the image.");
		}
		
		return embedData(sourceImage, dataList, colorChannelOffset);
	}
	
	/**
	 * The general one with colorChannelOffset = 3, used for reading in file
	 * instead of just a LinkedList of data
	 * @param sourceImage The image to have the data encrypted in
	 * @param data the data to be encrypted in file format.
	 * @return The encrypted image is returned.
	 */
	public static BufferedImage embedData(BufferedImage sourceImage, File data){
		return embedData(sourceImage, data, 3);
	}
	
	/**
	 * The general one which sets the colorChannelOffset to 3, no point using anything but
	 * 3, just left for easy recoding later (changing the base).
	 * @param sourceImage The image to have the data encrypted in
	 * @param data the data to be encrypted.
	 * @return The encrypted image is returned.
	 */
	public static BufferedImage embedData(BufferedImage sourceImage, LinkedList<Integer> data){
		return embedData(sourceImage, data, 3);
	}
	
	/**
	 * Embeds the data into the image
	 * @param sourceImage The image that the data should be stored in
	 * @param data the data in a LinkedList of intergers (byte values, between 0, and 256)
	 * @param colorChannelOffset the amount to alter a pixel plus or minus (3 = +-3
	 * @return the encrypted image.
	 */
	public static BufferedImage embedData(BufferedImage sourceImage, LinkedList<Integer> data, int colorChannelOffset){
		
		int[][] sourcePixels = ModifyImage.imageToPixels(sourceImage);

			//This turns the difference from a pixel into an integer value.
		RGBDifferenceToInteger rgbDiffToInt = new RGBDifferenceToInteger(colorChannelOffset);
			//This is used to convert between base 10 and the base that
			//the pixel alterations creates (+-1 = base3)
		BaseXToBase10 baseConversion = new BaseXToBase10(colorChannelOffset*2+1);
	
		int x=0, y=0;
		
		for(int i=0; i<data.size(); i++){
			
				//turn the character to an int
			int tempData = data.get(i);
			
			int pixelToBeModified = sourcePixels[x][y];
			
				//This makes sure the pixel does not have any
				//FF or 00 values. Change later to avoid single
				//values, not entire pixels.
			while(!isPixelModifiable(pixelToBeModified)){
					y++;
				if(y>=sourceImage.getWidth()){
					x++;
					y=0;
				}
				if( x > sourcePixels.length){
					System.err.println("Your image can not fit in this image." +
										" Try a larger Thumbnail or less black and white.");
					System.exit(1);
				}
				pixelToBeModified = sourcePixels[x][y];
			}
			
			int[] baseX = baseConversion.getBaseX(tempData);
			
			if(baseX.length > 3){
				System.err.println("Ut-oh, check embedData something is wrong with" +
						" your base 7 conversion.");
			}
			
				//This sets R, G and B (0-2) (since baseX == 3 hard coded)
			for(int color = 0; color < baseX.length; color++){
				pixelToBeModified = modifyRGorBPixel(baseX[color], pixelToBeModified, color, rgbDiffToInt);
			}
			
			sourcePixels[x][y] = pixelToBeModified;
			
			y++;
			if(y>=sourceImage.getWidth()){
				x++;
				y=0;
			}

			if(x>sourcePixels.length){
				System.err.println("Your image can not fit in this image." +
									" Try a larger Thumbnail or less black and white.");
				System.exit(1);
			}
		}
		
		/**
		 * The following sets the pixel after the last data pixel to 256,
		 * so that the end of the data can be found. (since the pixels can
		 * actually store up to 7^3 data (343)
		 */
		int[] tempBaseX = new int[3];			
		tempBaseX = baseConversion.getBaseX(256);
		for(int color = 0; color < tempBaseX.length; color++){
			sourcePixels[x][y] = modifyRGorBPixel(tempBaseX[color], sourcePixels[x][y], color, rgbDiffToInt);
		}
		
			//Returns the encrypted image
		return ModifyImage.pixelsToImage(sourcePixels);
	}
	

	/**
	 * This modifies a pixel based on the character and the
	 * character map.
	 * @param base10 a number in base 10, 0-26 to modify the pixel by
	 * @param pixel the pixel to embed the character into
	 * @return the modified pixel is returned, -1 is returned
	 * 			if the pixel can't be modified (IE FF or 00 in
	 * 			one of the positions needed to be modified.)
	 */
	public static int modifyRGorBPixel(int base10, int pixel, int RGB, RGBDifferenceToInteger rgbDiffToInt){
		if(RGB != R && RGB != G && RGB != B)
			throw new IllegalArgumentException("You must use a constant R, G, or B for RGB");
		
		int red = ModifyImage.getRed(pixel);
		int green = ModifyImage.getGreen(pixel);
		int blue = ModifyImage.getBlue(pixel);

		int valueToModify = rgbDiffToInt.getRGBDifferences(base10);
		
		if(RGB == R)
			red += valueToModify;
		else if(RGB == G)
			green += valueToModify;
		else if(RGB == B)
			blue += valueToModify;
		
		return ModifyImage.makePixel(red,green,blue);
	}
	
	
	/**
	 * Extracts the data from an image, using 3 as the colorChannelOffset
	 * @param sourceImage The original source image
	 * @param encryptedImage The image with the data encrypted in it
	 * @return The data is returned in a LinkedList of integers
	 */
	public static LinkedList<Integer> extractData(BufferedImage sourceImage,
										   BufferedImage encryptedImage){
		return extractData(sourceImage, encryptedImage, 3);
	}
	
	/**
	 * Extracts the data from an image.
	 * @param sourceImage The original source image
	 * @param encryptedImage The image with the data encrypted in it
	 * @param colorChannelOffset the amount that the pixels can differ by
	 * @return The data is returned in a LinkedList of integers
	 */
	public static LinkedList<Integer> extractData(BufferedImage sourceImage,
										   BufferedImage encryptedImage,
										   int colorChannelOffset){
		
		int[][] sourcePixels = ModifyImage.imageToPixels(sourceImage);
		int[][] encryptedPixels = ModifyImage.imageToPixels(encryptedImage);
		
			//This turns the difference from a pixel into an integer value.
		RGBDifferenceToInteger rgbDiffToInt = new RGBDifferenceToInteger(colorChannelOffset);
			//This is used to convert between base 10 and the base that
			//the pixel alterations creates (+-1 = base3)
		BaseXToBase10 baseConversion = new BaseXToBase10(colorChannelOffset*2+1);
		
		if(sourcePixels.length!=encryptedPixels.length){
			System.err.println("Source image != encrypted Image.");
		}
		
		LinkedList<Integer> dataBuffer = new LinkedList<Integer>();
		
		BreakAtMessageComplete:
			for(int y=0; y<encryptedPixels.length; y++){
				for(int x=0; x<encryptedPixels[y].length; x++){
				
				if(!isPixelModifiable(sourcePixels[y][x]))
					continue;
				
					//RG and B will be used (could be 2 RG, BR, GB.. etc) used for later
				int bitsToUse = 3;
				
				int temp[] = new int[bitsToUse];
				for(int colorChannels = 0; colorChannels < bitsToUse; colorChannels++){
					temp[colorChannels] = rgbDiffToInt.getInteger(
											differenceOfColorChannel(sourcePixels[y][x],
											 encryptedPixels[y][x], colorChannels));
				}
				int dataValue = baseConversion.getBase10(temp);
				
				if(dataValue<256)
					//convert the difference into base 10 and turn into character equivalent
					dataBuffer.add(dataValue);
				else
					//The data is done when the last pixel is some value above
					//255 (256).
					break BreakAtMessageComplete;
			}
		}
		
			//Returns the data from the image in an array List of integers
		return dataBuffer;
	}
	
	
	/**
	 * Turns the data (output from extract data) into a File object.
	 * @param data in a linked list of integers (0-255)
	 * @return the data in a File object.
	 */
	public static void dataListToFile(LinkedList<Integer> data, String filename){
		/**
		 * This creates the file to be decrypted from the image
		 */
		File fileToBeDecrypted;
		FileOutputStream fout = null;
		try{
			fileToBeDecrypted = new File(filename);
			fout = new FileOutputStream(fileToBeDecrypted);
		}catch(Exception e){
			System.err.println("There was a problem with the file being encrypted.");
		}
		
		/**
		 * This adds all the data from to file from a LinkedList
		 * byte by byte.
		 */
		try{
			for(int i=0; i<data.size();i++)
				fout.write(data.get(i));
		}catch(IOException e){
			System.err.println("There was an error with your file to be encrypted in the image.");
		}
	}
	

	/**
	 * Returns the integer value (0-26) based on the difference
	 * 		of the two pixels.
	 * @param sourcePixels the original pictures pixels
	 * @param encryptedPixels the encrypted pictures pixels
	 * @param colorChannel is one of the finals r, g or B (0, 1, or 2)
	 * @return the difference of the pixels
	 */
	public static int differenceOfColorChannel(int sourcePixels, int encryptedPixels, int colorChannel){
		
		if(!isPixelModifiable(sourcePixels))
			throw new IllegalArgumentException("Pixel was not and can not be modified upper limit");

		if(colorChannel == 0){
			int red = ModifyImage.getRed(sourcePixels);
			return ModifyImage.getRed(encryptedPixels) - red;
		}
		else if(colorChannel == 1){
			int green = ModifyImage.getGreen(sourcePixels);
			return ModifyImage.getGreen(encryptedPixels) - green;
		}
		else if(colorChannel == 2){
			int blue = ModifyImage.getBlue(sourcePixels);
			return (ModifyImage.getBlue(encryptedPixels) - blue);
		}
		else{
				//Some other colorChannel was input
			throw new IllegalArgumentException("The finals R, G or B must be passed in" +
													" the color channel variable");
		}
	}
	
	
	/**
	 * This checks if a pixel is close to FF or 00 so that
	 * modifying it would go above FF or below 00.
	 * @param pixel the pixel to be checked.
	 * @return false is returned if the pixel contains any values
	 * 			at FF (assuming amount per channel to alter is 1)
	 * 			(FE if it's 2, FD if it's 3... etc)
	 */
	private static boolean isPixelModifiable(int pixel){
		
		int colorChannelOffset = 3; // = amtChannelAlter;

		int red = ModifyImage.getRed(pixel);
		int green = ModifyImage.getGreen(pixel);
		int blue = ModifyImage.getBlue(pixel);
		
		if( !isColorChannelModifiable(red,colorChannelOffset) ||
			!isColorChannelModifiable(green,colorChannelOffset)  ||
			!isColorChannelModifiable(blue,colorChannelOffset) )
			return false;
		
		return true;
	}
	
	/**
	 * This checks if a color channel is close to FF or 00 so that
	 * modifying it would go above FF or below 00.
	 * @param colorChannel the color channel to be checked.
	 * @param colorChanOffset The amount that the color channel will be changed bye (1 for +-1, 3 for +- 3
	 * @return false is returned if the colorChannel is at
	 * 			 FF (assuming amount per channel to alter is 1)
	 * 			(FE if it's 2, FD if it's 3... etc)
	 */
	private static boolean isColorChannelModifiable(int colorChannel, int colorChanOffset){		
		if( colorChannel >= 0xFF-(colorChanOffset-1)    || colorChannel <= 0x00+(colorChanOffset-1))
			return false;
		
		return true;
	}
	
	/**
	 * This finds the smallest width of the image that data of size <i>dataSize</i>
	 * can fit in
	 * @param srcImage The source image.
	 * @param
	 * @return the smallest estimated width required.
	 */
	public static int estimateSmallestThumbnailPossible(BufferedImage srcImage, int dataSize){
		double tempWidth=1, tempHeight=1;
		
		double ratio = (double)findNumModifiablePixels(srcImage, 3)/(double)(srcImage.getWidth()*srcImage.getHeight());
		
		while(((ratio)*(tempWidth*tempHeight)) < dataSize){
			tempWidth+=1.0;
			tempHeight = Math.floor(tempWidth * 
						((double)srcImage.getHeight()/srcImage.getWidth()));
		}
		
		return (int)Math.ceil(tempWidth);
	}
	
	/**
	 * This method finds all pixels that are modifiable
	 * @param srcImage image to look at the pixels.
	 * @param colorChannelOffset (1 = +-1, 2 = +-2...)
	 * @return the total number of pixels that are modifiable.
	 */
	public static int findNumModifiablePixels(BufferedImage srcImage, int colorChannelOffset){
		int[][] image = ModifyImage.imageToPixels(srcImage);
		
		int total = 0;
		
		for(int y=0; y<image.length; y++){
			for(int x=0; x<image[y].length; x++){
					if(!isColorChannelModifiable(ModifyImage.getRed(image[y][x]),3))
						continue;
					if(!isColorChannelModifiable(ModifyImage.getGreen(image[y][x]),3))
						continue;
					if(!isColorChannelModifiable(ModifyImage.getBlue(image[y][x]),3))
						continue;
					total++;
			}
		}
		
		return total;
	}
	
	
}
